/*
 * "$Id: image.c,v 1.5 93/12/05 10:42:39 mike Exp Locker: mike $"
 *
 * Image manipulation routines for TOPCL, an SGI image to PCL converter.
 *
 * Copyright (c) 1993 Easy Software Products
 *                    P.O. Box 256
 *                    Park Hall, MD  20667
 *                    Attn: Michael Sweet (dodgecolt@delphi.com)
 *
 * This file is part of TOPCL.
 *
 * TOPCL is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 * to anyone for the consequences of using it or for whether it serves any
 * particular purpose or works at all, unless he says so in writing.  Refer
 * to the TOPCL General Public License for full details.
 *
 * Everyone is granted permission to copy, modify and redistribute
 * TOPCL, but only under the conditions described in the TOPCL
 * General Public License.  A copy of this license is supposed to have been
 * given to you along with TOPCL so you can know your rights and
 * responsibilities.  It should be in a file named COPYING.  Among other
 * things, the copyright notice and this notice must be preserved on all
 * copies.
 *
 * Contents:
 *
 *  get_gamma()   - Get the system gamma value.
 *  AdjustImage() - Gamma-correct an image.
 *  AllocImage()  - Allocate memory for an image.
 *  FreeImage()   - Free memory for an image.
 *  GreyImage()   - Convert an image to greyscale.
 *  LoadImage()   - Load an RGB image from disk.
 *
 * Revision History:
 *
 *  $Log:	image.c,v $
 * Revision 1.5  93/12/05  10:42:39  mike
 * Updated documentation for v1.1 release.
 * 
 * Revision 1.4  93/12/05  02:22:26  mike
 * Updated color correction code for deskjets and paper media output.
 * Changed LoadImage() code to use 'getrow()' because the 'getpix()' macro is
 *   broken.
 * 
 * Revision 1.3  93/06/10  22:34:08  mike
 * Cleaned up for version 1.0 release.
 * 
 * Revision 1.2  93/06/09  23:18:11  mike
 * Speeded gamma correction code.
 * 
 * Revision 1.1  93/05/24  19:41:55  mike
 * Initial revision
 * 
 *
 */

/*
 * Include necessary headers.
 */

#include "topcl.h"
#include <math.h>

#ifdef DEBUG
# include <gl/gl.h>
#endif /* DEBUG */

/*
 * Macros from <macros.h>; we can't include <macros.h> because it also
 * defines DEBUG, one of our flags to insert various debugging code.
 */

#define max(a,b) ((a)<(b) ? (b) : (a))
#define min(a,b) ((a)>(b) ? (b) : (a))
#define abs(x)   ((x)>=0 ? (x) : -(x))

/*
 * Color correction constant for color deskjets.  This is used in
 * 'AdjustColors()' to increase the red output level slightly to compensate
 * for the cyan ink used.
 */

#define ADJUST   12


/*
 * 'get_gamma()' - Get the system default gamma value...
 */

static float
get_gamma(void)
{
 FILE  *gamma_cmd; /* Pipe file to gamma command */
 float gammaval;   /* Gamma value from gamma command */


/*
 * Set the default gamma correction value (1.0) and try reading the output
 * from the gamma(6D) command...
 */

 gammaval = 1.0;

 if ((gamma_cmd = popen("/usr/sbin/gamma", "r")) != NULL)
  {
   fscanf(gamma_cmd, "%f", &gammaval);
   pclose(gamma_cmd);
  };

 return (gammaval);
}


/*
 * 'AdjustImage()' - Perform gamma correction on an image.
 */

void
AdjustImage(RGB_IMAGE *i,   /* I - Image to adjust */
            float gammaval) /* I - Gamma correction value */
{
 int           rgb,         /* RGB component, 0 - 2 */
               val;         /* Looping var... */
 RGB_PIXEL     *current,    /* Current pixel */
               *last;       /* Last pixel in image */
 float         igammaval;   /* 1/gammaval */
 short         lut[256];    /* Look-up table for correction values... */
 int           cmin,        /* Minimum color value */
               cmax,        /* Maximum color value */
               crange,      /* Range of color values */
               cred;        /* New red color value */
#ifdef DEBUG
 int           x,y;         /* Image position */
#endif /* DEBUG */


 if (i == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to adjust non-existent image!\n");
   abort();
  };

 if (i->pixels == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to adjust damaged image!\n");
   abort();
  };

/*
 * If no gamma value has been given, get the system default gamma and use it.
 * (I wish SGI had thought to put the gamma value in the image headers!)
 */

 if (gammaval <= 0.0)
  gammaval = get_gamma();
 if (gammaval <= 0.0)
  gammaval = 1.0;

 igammaval = 1.0 / gammaval;

#ifdef DEBUG
 fprintf(stderr, "gammaval = %f, igammaval = %f\n", gammaval, igammaval);
#endif /* DEBUG */

/*
 * Generate a gamma correction table for each of 256 values...
 */

 for (val = 0; val < 256; val ++)
  {
   lut[val] = (int)(255.0 * pow(val / 255.0, igammaval) + 0.5);

   if (OutputMedia == PCL_PAPER)
    {
    /*
     * Adjust brightness for paper as less light is transmitted from paper
     * than through a transparency.  Clamp values to a range of 0 to 255
     * to avoid distortion.
     */

     lut[val] += 16;
     if (lut[val] > 255)
      lut[val] = 255;
    };
  };

/*
 * Replace each pixel in the image with it's corrected brother...
 */

#ifdef DEBUG
 x = 0;
 y = 0;
#endif /* DEBUG */

 for (current = i->pixels, last = i->pixels + i->width * i->height;
      current < last;
      current ++)
  {
   (*current)[0] = lut[(*current)[0]];
   (*current)[1] = lut[(*current)[1]];
   (*current)[2] = lut[(*current)[2]];

   if (OutputType == PCL_COLOR)
    {
    /*
     * Correct inkjet colors by adding a little red.  This yields brighter
     * colors that match the true colors much more closely.  By adding red
     * we actually remove some cyan; colors otherwise appear a bit greenish.
     */

     cmin   = min(min((*current)[0], (*current)[1]), (*current)[2]);
     cmax   = max(max((*current)[0], (*current)[1]), (*current)[2]);
     crange = cmax - cmin + 1;

     cred = (*current)[0] * (cmax * (ADJUST + 1)) / (crange + cmax * ADJUST);
     if (cred > 255)
      (*current)[0] = 255;
     else
      (*current)[0] = cred;
    };

#ifdef DEBUG
  /*
   * Draw the updated color cell on the screen.
   */

   RGBcolor((*current)[0], (*current)[1], (*current)[2]);
   pnt2i(x, y);

   x ++;
   if (x >= i->width)
    {
     x = 0;
     y ++;
    };
#endif /* DEBUG */
  };
}


/*
 * 'AllocImage()' - Allocate memory for an image.  Exit on errors.
 *                  Returns a pointer to an RGB_IMAGE structure.
 */

RGB_IMAGE *
AllocImage(int width,  /* I - Pixel width of image */
           int height) /* I - Pixel height of image */
{
 RGB_IMAGE *i;         /* RGB image structure pointer */


 if (width == 0 || height == 0)
  {
   fprintf(stderr, "topcl: Illegal image dimensions %dx%d!\n", width, height);
   exit(EINVAL);
  };

 if ((i = (RGB_IMAGE *)malloc(sizeof(RGB_IMAGE))) == NULL)
  {
   fprintf(stderr, "topcl: Allocation of %dx%d image buffer failed - %s\n",
           width, height, strerror(errno));
   exit(errno);
  };

 i->width  = width;
 i->height = height;

 if ((i->pixels = (RGB_PIXEL *)calloc(width * height, sizeof(RGB_PIXEL))) == NULL)
  {
   free(i);

   fprintf(stderr, "topcl: Allocation of %dx%d image buffer failed - %s\n",
           width, height, strerror(errno));
   exit(errno);
  };

 return (i);
}


/*
 * 'FreeImage()' - Free an image.
 */

void
FreeImage(RGB_IMAGE *i) /* I - Image structure pointer */
{
 if (i == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to free non-existent image!\n");
   abort();
  };

 if (i->pixels == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to free damaged image!\n");
   abort();
  };

 free(i->pixels);
 free(i);
}


/*
 * 'GreyImage()' - Convert an RGB image to greyscale...
 */

void
GreyImage(RGB_IMAGE *i) /* I - RGB image structure */
{
 RGB_PIXEL *current,    /* Current pixel */
           *last;       /* Last pixel in image */


 if (i == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to greyscale non-existent image!\n");
   abort();
  };

 if (i->pixels == NULL)
  {
   fprintf(stderr, "topcl: Program error; attempted to greyscale damaged image!\n");
   abort();
  };

 for (current = i->pixels, last = i->pixels + i->width * i->height;
      current < last;
      current ++)
 /*
  * Do conversion from RGB to luminance (greyscale) using NTSC constants.
  * Greyscale values are put the last plane for ease of use.
  */

  (*current)[2] = ((*current)[0] * 30 +       /* [RED * 0.30] */
                   (*current)[1] * 59 +       /* [GREEN * 0.59] */
                   (*current)[2] * 11) / 100; /* [BLUE * 0.11] */
}


/*
 * 'LoadImage()' - Load an SGI image file and return a pointer to the RGB
 *                 image structure containing the loaded image.
 *
 * This code looks so convoluted because it is.  SGI images can come in as
 * greyscale, interleaved RGB, or colormap.  Each type of file gets read in
 * by a different section of code...
 *
 * Presently, colormap images are not supported.
 *
 */

RGB_IMAGE *
LoadImage(char *filename) /* I - Name of file to load */
{
 IMAGE         *ip;       /* SGI image pointer */
 RGB_IMAGE     *i;        /* RGB image pointer */
 RGB_PIXEL     *pixel;    /* Current RGB pixel... */
 int           x, y;      /* Current X & Y in image */
 int           rgb;       /* Current RGB plane */
 short         *row;      /* Data for a row */


/*
 * First try opening the image file.  If it fails, print an error message
 * and exit...
 */

 if ((ip = iopen(filename, "r", 0, 0, 0, 0, 0)) == NULL)
  {
   fprintf(stderr, "topcl: Unable to open \'%s\' for reading - %s\n",
           filename, strerror(errno));
   exit(errno);
  };

 if (ip->colormap != CM_NORMAL)
  {
   fprintf(stderr, "topcl: Sorry, colormap images like \'%s\' are not supported!\n",
           filename);
   iclose(ip);
   exit(EINVAL);
  };

/*
 * Next try allocating memory for the RGB image...
 */

 i = AllocImage(ip->xsize, ip->ysize);

#ifdef DEBUG
/*
 * Open a debug window to show the image.
 */

 prefsize(ip->xsize, ip->ysize);
 foreground();
 winopen(filename);
 RGBmode();
 gconfig();

 viewport(0, ip->xsize - 1, 0, ip->ysize - 1);
 ortho2(0, ip->xsize - 1, ip->ysize - 1, 0);
#endif /* DEBUG */

/*
 * Based upon the type of image, read each pixel row in the file.  Because of
 * bugs in the 'getpix()' macro, we MUST read a whole row of pixels at a time.
 * Otherwise some images are not read properly.  On a happy note, this speeds
 * image loading by an order of magnitude.
 */

 if ((row = (short *)calloc(ip->xsize, sizeof(short))) == NULL)
  {
   fprintf(stderr, "topcl: Unable to allocate image buffer - %s\n",
           strerror(errno));
   exit(errno);
  };
  
 if (BPP(ip->type) == 1 && ip->zsize == 1)
  {
  /*
   * This is a greyscale image.  Loading is easy...
   */

   OutputType = PCL_BW; /* Force black-and-white output */

   for (pixel = i->pixels + (ip->ysize - 1) * ip->xsize, y = ip->ysize - 1;
        y >= 0;
        y --, pixel -= 2 * ip->xsize)
    {
     getrow(ip, row, ip->ysize - 1 - y, 0);

     for (x = 0; x < ip->xsize; x ++, pixel ++)
      (*pixel)[0] = 
      (*pixel)[1] = 
      (*pixel)[2] = row[x];
    };
  }
 else if (BPP(ip->type) == 1 && ip->zsize == 3)
  {
  /*
   * This is an interlaced RGB image...
   */

   for (rgb = 0; rgb < 3; rgb ++)
    for (pixel = i->pixels + (ip->ysize - 1) * ip->xsize, y = ip->ysize - 1;
         y >= 0;
         y --, pixel -= 2 * ip->xsize)
     {
      getrow(ip, row, ip->ysize - 1 - y, rgb);

      for (x = 0; x < ip->xsize; x ++, pixel ++)
       (*pixel)[rgb] = row[x];
     };
  }
 else
  {
  /*
   * This is some other image format we don't understand.  Tell the user
   * and exit...
   */

   fprintf(stderr, "topcl: \'%s\' is in a format I don\'t understand:\n",
           filename);
   fprintf(stderr, "       xsize = %d, ysize = %d, zsize = %d\n",
           ip->xsize, ip->ysize, ip->zsize);
   fprintf(stderr, "       type = %04X\n", ip->type);

   iclose(ip);
   FreeImage(i);
   free(row);
   exit(EINVAL);
  };

#ifdef DEBUG
/*
 * Show the image on the screen.
 */

 ringbell();

 for (y = 0, pixel = i->pixels; y < ip->ysize; y ++)
  for (x = 0; x < ip->xsize; x ++, pixel ++)
   {
    RGBcolor((*pixel)[0], (*pixel)[1], (*pixel)[2]);
    pnt2i(x, y);
   };
#endif /* DEBUG */

/*
 * Free the row buffer, close the image file, and return the image we've
 * loaded.
 */

 free(row);
 iclose(ip);

 return (i);
}


/*
 * End of "$Id: image.c,v 1.5 93/12/05 10:42:39 mike Exp Locker: mike $".
 */

