/*
 * "$Id: output.c,v 1.5 93/12/05 10:42:53 mike Exp Locker: mike $"
 *
 * Output routines for TOPCL, an SGI image to PCL converter.
 *
 * Copyright (c) 1993 Easy Software Products
 *                    P.O. Box 256
 *                    Park Hall, MD  20667
 *                    Attn: Michael Sweet (dodgecolt@delphi.com)
 *
 * This file is part of TOPCL.
 *
 * TOPCL is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
 * to anyone for the consequences of using it or for whether it serves any
 * particular purpose or works at all, unless he says so in writing.  Refer
 * to the TOPCL General Public License for full details.
 *
 * Everyone is granted permission to copy, modify and redistribute
 * TOPCL, but only under the conditions described in the TOPCL
 * General Public License.  A copy of this license is supposed to have been
 * given to you along with TOPCL so you can know your rights and
 * responsibilities.  It should be in a file named COPYING.  Among other
 * things, the copyright notice and this notice must be preserved on all
 * copies.
 *
 * Contents:
 *
 *  burkes_dither() - Dither pixel error using Burke's coefficients.
 *  output_line()   - Output a line of graphics.
 *  OutputImage()   - Output an image.
 *
 * Revision History:
 *
 *  $Log:	output.c,v $
 * Revision 1.5  93/12/05  10:42:53  mike
 * Change output code to set depletion and shingling for both color and
 *   black-and-white inkjets.
 * Updated documentation for v1.1 release.
 * 
 * Revision 1.4  93/12/05  02:23:38  mike
 * Added mode 9 compression code.
 * Removed code to output every-other dot on paper.
 * Made the 'burkes_dither()' function an inline.
 * 
 * Revision 1.3  93/06/10  22:34:20  mike
 * Cleaned up for version 1.0 release.
 * 
 * Revision 1.2  93/06/09  23:18:39  mike
 * Added mode 3 compression code.
 * 
 * Revision 1.1  93/05/24  19:42:07  mike
 * Initial revision
 * 
 *
 */

/*
 * Include necessary headers.
 */

#include "topcl.h"


/*
 * 'burkes_dither()' - Dither pixel error using Burke's coefficients.
 */

static void
burkes_dither(short *p0,   /* I - Current dot line */
              short *p1,   /* I - Next dot line */
              int   error) /* I - Amount of error to distribute */
{
/*
 *           X    1/4  1/8
 * 1/16 1/8  1/4  1/8  1/16
 *
 */

 if (error < 0)
  error += 3; /* Avoid division errors with bitshifts... */

 error >>= 2;

 *(p0 + 1) += error;
 *(p1 + 0) += error;

 error >>= 1;

 *(p0 + 2) += error;
 *(p1 - 1) += error;
 *(p1 + 1) += error;

 error >>= 1;

 *(p1 - 2) += error;
 *(p1 + 2) += error;
}


/*
 * 'output_line()' - Optionally compress and output a line of raster graphics.
 *
 * This starts at the highest supported compression on the printer and then
 * drops back to the lower ones if compression yields more output bytes.
 */

#define MAX_OUTBYTES 600 /* Maximum bytes-per-line, 8 inch line, 600 dpi */

static unsigned char comp_seed[3][MAX_OUTBYTES];
                                  /* Mode 3 seed buffer */
static unsigned char comp_buffer[3 * MAX_OUTBYTES];
                                  /* Compression buffer */
static int           last_mode = -1;
                                  /* Last compression mode used */

static void
output_line(unsigned char *line,  /* I - Line of GFX to print */
            int           length, /* I - Length of GFX in bytes */
            int           plane)  /* I - RGB plane for color output, 2 otherwise */
{
 unsigned char *line_ptr,         /* Current byte pointer */
               *line_end,         /* End-of-line byte pointer */
               *comp_ptr,         /* Pointer into compression buffer */
               *start,            /* Start of sequence (for compression) */
               *seed;             /* Seed data pointer for mode 3,9 compression */
 int           i,                 /* Looping var */
               count,             /* Count of bytes for output */
               offset,            /* Offset for mode 3,9 compression */
               comp_mode;         /* Compression mode finally used... */


 switch (OutputMode)
  {
   case PCL_MODE9 :
      /*
       * Do DeskJet mode 9 compression.  This is similar to mode 3 but can
       * do RLE as well...
       */

       line_ptr = line;
       line_end = line + length;

       comp_mode = 9;

       comp_ptr = comp_buffer;
       seed     = comp_seed[plane];

       while (line_ptr < line_end &&
              (comp_ptr - comp_buffer) < length)
        {
        /*
         * Find the next non-matching sequence...
         */

         start = line_ptr;
         while (*line_ptr == *seed &&
                line_ptr < line_end)
          {
           line_ptr ++;
           seed ++;
          };

         if (line_ptr == line_end)
          break;

         offset = line_ptr - start;

        /*
         * Find non-matching bytes...
         */

         start = line_ptr;
         count = 0;
         while ((*line_ptr != *seed || count < 4) &&
                line_ptr[0] != line_ptr[1] &&
                line_ptr < line_end)
          {
           line_ptr ++;
           seed ++;
           count ++;
          };

        /*
         * Place non-RLE compression data in the buffer; see HP manuals
         * for details...
         */

         if (count > 0)
          {
          /*
           * Set initial count of bytes.
           */

           if (count > 7)
            *comp_ptr = 7;
           else
            *comp_ptr = count - 1;

           if (offset >= 15)
            {
            /*
             * Output multi-byte offset...
             */

             *comp_ptr++ |= 0x78;

             offset -= 15;
             while (offset >= 255)
              {
               *comp_ptr++ = 255;
               offset      -= 255;
              };

             *comp_ptr++ = offset;
            }
           else
            {
            /*
             * Output single-byte offset...
             */

             *comp_ptr++ |= (offset << 3);
            };

           if (count > 7)
            {
            /*
             * Output multi-byte count...
             */

             i = count - 8;
             while (i >= 255)
              {
               *comp_ptr++ = 255;
               i           -= 255;
              };

             *comp_ptr++ = i;
            };

           memcpy(comp_ptr, start, count);
           comp_ptr += count;
           offset = 0;
          };

        /*
         * Find matching (repeated) bytes...
         */

         if (*line_ptr != *seed &&
             line_ptr[0] == line_ptr[1] &&
             line_ptr < line_end)
          {
           start = line_ptr;
           count = 0;
           while ((*line_ptr != *seed || count < 4) &&
                  line_ptr[0] == line_ptr[1] &&
                  line_ptr < line_end)
            {
             line_ptr ++;
             seed ++;
             count ++;
            };

          /*
           * Place RLE compression data in the buffer; see HP manuals
           * for details...
           */

          /*
           * Set initial count of bytes.
           */

           if (count > 31)
            *comp_ptr = 0x80 | 31;
           else
            *comp_ptr = 0x80 | (count - 1);

           if (offset >= 3)
            {
            /*
             * Output multi-byte offset...
             */

             *comp_ptr++ |= 0x60;

             offset -= 3;
             while (offset >= 255)
              {
               *comp_ptr++ = 255;
               offset      -= 255;
              };

             *comp_ptr++ = offset;
            }
           else
            {
            /*
             * Output single-byte offset...
             */

             *comp_ptr++ |= (offset << 5);
            };

           if (count > 31)
            {
            /*
             * Output multi-byte count...
             */

             i = count - 32;
             while (i >= 255)
              {
               *comp_ptr++ = 255;
               i            -= 255;
              };

             *comp_ptr++ = i;
            };

           *comp_ptr++ = *line_ptr++;
           seed ++;
          };
        };

      /*
       * Check to see if compression was better...
       */

       line_ptr = comp_buffer;
       line_end = comp_ptr;
       if ((line_end - line_ptr) < length)
        break; /* Yep */

      /*
       * Drop back to mode 3 (delta-row) compression...
       */

   case PCL_MODE3 :
      /*
       * Do delta-row compression...
       */

       line_ptr = line;
       line_end = line + length;

       comp_mode = 3;

       comp_ptr = comp_buffer;
       seed     = comp_seed[plane];

       while (line_ptr < line_end &&
              (comp_ptr - comp_buffer) < length)
        {
        /*
         * Find the next non-matching sequence...
         */

         start = line_ptr;
         while (*line_ptr == *seed &&
                line_ptr < line_end)
          {
           line_ptr ++;
           seed ++;
          };

         if (line_ptr == line_end)
          break;

         offset = line_ptr - start;

        /*
         * Find up to 8 non-matching bytes...
         */

         start = line_ptr;
         count = 0;
         while (*line_ptr != *seed &&
                line_ptr < line_end &&
                count < 8)
          {
           line_ptr ++;
           seed ++;
           count ++;
          };

        /*
         * Place mode 3 compression data in the buffer; see HP manuals
         * for details...
         */

         if (offset >= 31)
          {
          /*
           * Output multi-byte offset...
           */

           *comp_ptr++ = ((count - 1) << 5) | 31;

           offset -= 31;
           while (offset >= 255)
            {
             *comp_ptr++ = 255;
             offset    -= 255;
            };

           *comp_ptr++ = offset;
          }
         else
          {
          /*
           * Output single-byte offset...
           */

           *comp_ptr++ = ((count - 1) << 5) | offset;
          };

         memcpy(comp_ptr, start, count);
         comp_ptr += count;
        };

      /*
       * Check to see if compression was better...
       */

       line_ptr = comp_buffer;
       line_end = comp_ptr;
       if ((line_end - line_ptr) < length)
        break; /* Yep */

      /*
       * Drop back to mode 2 (TIFF RLE) compression...
       */

   case PCL_MODE2 :
      /*
       * Do TIFF RLE compression...
       */

       line_ptr = line;
       line_end = line + length;

       comp_mode = 2;

       comp_ptr = comp_buffer;

       while (line_ptr < line_end &&
              (comp_ptr - comp_buffer) < length)
        if ((line_ptr + 1) >= line_end)
         { /* Single byte... */
          *comp_ptr++ = 0;
          *comp_ptr++ = *line_ptr++;
         }
        else if (*line_ptr == *(line_ptr + 1))
         { /* Repeated sequence... */
          line_ptr ++;
          count = 255;

          while (*line_ptr == (*line_ptr + 1) &&
                 count > 128 &&
                 line_ptr < line_end)
           {
            line_ptr ++;
            count --;
           };

          *comp_ptr++ = count;
          *comp_ptr++ = *line_ptr++;
         }
        else
         { /* Non-repeated sequence... */
          start    =  line_ptr;
          line_ptr += 2;
          count    =  2;

          while (*line_ptr != *(line_ptr + 1) &&
                 count < 127 &&
                 line_ptr < line_end)
           {
            line_ptr ++;
            count ++;
           };

          *comp_ptr++ = count - 1;

          memcpy(comp_ptr, start, count);
          comp_ptr += count;
         };

      /*
       * Check to see if compression was better...
       */

       line_ptr = comp_buffer;
       line_end = comp_ptr;

       if ((line_end - line_ptr) < length)
        break; /* Yep */

      /*
       * Drop back to mode 1 RLE compression...
       */

   case PCL_MODE1 :
      /*
       * Skip RLE compression; this is a simpler form of mode 2 and won't
       * do as well as a result...
       */

      /*
       * Drop back to mode 0 (no) compression...
       */

   case PCL_MODE0 :
      /*
       * Do no compression...
       */

       line_ptr = line;
       line_end = line + length;

       comp_mode = 0;
       break;
  };

/*
 * Set the graphics compression mode if necessary and output a line of
 * graphics...
 */

 if (OutputMode > PCL_MODE0 && comp_mode != last_mode)
  {
   last_mode = comp_mode;
   printf("%dm%d%c", comp_mode,               /* Set the compression mode */
          line_end - line_ptr, "vvw"[plane]); /* and length of data */
  }
 else
  printf("%d%c", line_end - line_ptr, "vvw"[plane]);
                                              /* Set data length */

 if (OutputMode >= PCL_MODE3)
  memcpy(comp_seed[plane], line, length);     /* Update the seed row... */

 fwrite(line_ptr, line_end - line_ptr, sizeof(unsigned char), stdout);
}


/*
 * 'OutputImage()' - Output PCL for one of HPs many printers...
 */

void
OutputImage(RGB_IMAGE *i,       /* I - RGB image to output */
            int       left,     /* I - Leftmost pixel to display */
            int       top,      /* I - Topmost pixel to display */
            int       right,    /* I - Rightmost pixel to display */
            int       bottom)   /* I - Bottommost pixel to display */
{
 int           x,              /* Horizontal position in pbuf... */
               y,              /* Vertical position in pbuf... */
               istep,          /* Integral step along x axis... */
               idir,           /* Direction along x axis... */
               imod,           /* Remainder from scaling division... */
               ierror,         /* Total accumulated division error... */
               iwidth,         /* Width multiplier... */
               iheight,        /* Height multiplier... */
               ileft,          /* Adjusted (rotated/flipped) sides */
               iright,
               itop,
               ibottom,
               iy,             /* Image vertical position... */
               w,              /* Width of region to print... */
               h,              /* Height of region to print... */
               rgb,            /* R, G, or B being printed... */
               pixel,          /* Current adjusted pixel... */
               error,          /* Brightness error... */
               odd;            /* 0 on even lines, 1 on odd lines */
 short         *p0, *p1,       /* Dither pointers... */
               *pbuf[2][3];    /* Line dither buffers... */
 RGB_PIXEL     *iline;         /* Pointer to start of this image line... */
 unsigned char *lbyte,         /* Output line byte... */
               *lbuf;          /* Output line buffer... */
 int           lbit,           /* Output line bit... */
               lwidth,         /* Byte width of printer buffer */
               printer_width,  /* Width of the printer image... */
               printer_height; /* Height of the printer image... */


/*
 * Compute printed image width and height...
 */

 if (right < left)
  {
   x     = right;
   right = left;
   left  = x;
  };

 if (bottom < top)
  {
   y      = bottom;
   bottom = top;
   top    = y;
  };

 w = right - left + 1;
 h = bottom - top + 1;

/*
 * Compute printer dot width/height for the selected orientation...
 */

 if (OutputOrient == PCL_PORTRAIT)
  {
   printer_width  = PaperWidth * OutputDpi / 720;
   printer_height = printer_width * h / w;

   if (printer_height > (PaperHeight * OutputDpi / 720))
    {
     printer_height = PaperHeight * OutputDpi / 720;
     printer_width  = printer_height * w / h;
    };

   lwidth  = (printer_width + 7) / 8;

   ileft   = left;
   iright  = right;
   itop    = top;
   ibottom = bottom;

   iwidth  = i->width;
   iheight = h;
   idir    = 1;

   istep   = idir * (w / printer_width);
   imod    = w % printer_width;
  }
 else
  {
   printer_width  = PaperWidth * OutputDpi / 720;
   printer_height = printer_width * w / h;

   if (printer_height > (PaperHeight * OutputDpi / 720))
    {
     printer_height = PaperHeight * OutputDpi / 720;
     printer_width  = printer_height * h / w;
    };

   lwidth  = (printer_width + 7) / 8;

   iright  = top * i->width;
   ileft   = bottom * i->width;
   ibottom = right;
   itop    = left;

   iwidth  = 1;
   iheight = w;
   idir    = (-i->width);

   istep   = idir * (h / printer_width);
   imod    = h % printer_width;
  };

 if (OutputFlip & PCL_FLIP_X)
  {
   x      = ileft;
   ileft  = iright;
   iright = x;

   idir   = (-idir);
   istep  = (-istep);
  };

 if (OutputFlip & PCL_FLIP_Y)
  {
   y       = itop;
   itop    = ibottom;
   ibottom = y;

   iheight = (-iheight);
  };

#ifdef DEBUG
 fprintf(stderr, "iwidth = %d, idir = %d, istep = %d, imod = %d\n",
         iwidth, idir, istep, imod);
 fprintf(stderr, "itop = %d, ibottom = %d, ileft = %d, iright = %d\n",
         itop, ibottom, ileft, iright);
#endif /* DEBUG */

/*
 * Allocate buffers...
 */

 for (rgb = OutputType; rgb < 3; rgb ++)
  {
   pbuf[0][rgb] = (short *)calloc(printer_width + 4, sizeof(short));
   pbuf[1][rgb] = (short *)calloc(printer_width + 4, sizeof(short));

   if (pbuf[0][rgb] == NULL || pbuf[1][rgb] == NULL)
    {
     while (rgb >= OutputType)
      {
       if (pbuf[0][rgb] != NULL)
        free(pbuf[0][rgb]);
       if (pbuf[1][rgb] != NULL)
        free(pbuf[1][rgb]);

       rgb --;
      };

     return;
    };
  };

 lbuf = (unsigned char *)calloc(lwidth, sizeof(unsigned char));

 if (lbuf == NULL)
  {
   for (rgb = OutputType; rgb < 3; rgb ++)
    {
     free(pbuf[0][rgb] - 2);
     free(pbuf[1][rgb] - 2);
    };

   return;
  };

#ifdef DEBUG
 fprintf(stderr, "Image (%d x %d), printer (%d, %d)\n",
         i->width, i->height, printer_width, printer_height);
#endif /* DEBUG */

/*
 * Initialize the printer for graphics...
 */

 if (OutputType == PCL_COLOR)
  printf("\033*r-3U");  /* Do CMY color GFX */

 if (PrinterModel >= PCL_DJ500)
  printf("\033*o2d2Q"); /* Shingle and deplete as much as possible */

 printf("\033*p%dx%dY",               /* Center the graphics */
        (30 * PaperWidth / 72 - printer_width * 300 / OutputDpi) / 2,
        (30 * PaperHeight / 72 - printer_height * 300 / OutputDpi) / 2);
 printf("\033*r1A");                  /* Start graphics at current position */
 printf("\033*b");                    /* All graphics commands start with *b */

 memset(comp_seed, 0, sizeof(comp_seed));
                                      /* Clear the compression seed row... */

/*
 * Dither each output line; note that we are NOT using a serpentine scan, as
 * it tends to 'texture' the output with Burke's coefficients.  While
 * interesting to look at, it isn't acceptable.
 */

 for (y = 0, odd = 0; y < printer_height; y ++, odd ^= 1)
  {
  /*
   * Compute image y position
   */

   iy = itop + iheight * y / printer_height;

  /*
   * Do each plane (3 for RGB, 1 for BW).
   */

   for (rgb = OutputType; rgb < 3; rgb ++)
    {
     iline = i->pixels + iy * iwidth + ileft;

    /*
     * Clear the old second line and output buffer...
     */

     memset(pbuf[odd ^ 1][rgb] + 2, 0, sizeof(short) * printer_width);
     memset(lbuf, 0, sizeof(unsigned char) * lwidth);

     for (ierror = 0, x = printer_width - 1, lbyte = lbuf, lbit = 128,
            p0 = pbuf[odd][rgb] + 2, p1 = pbuf[odd ^ 1][rgb] + 2;
          x > 0;
          x --, p0 ++, p1 ++)
      {
      /*
       * Compute the net pixel brightness and brightness error.  Set a dot
       * if necessary...
       */

       pixel = (*iline)[rgb] + *p0;

       if (pixel > 128)
        {
         /* Already 0'd... */
         error = pixel - 255;
        }
       else
        {
         *lbyte |= lbit;
         error = pixel;
        };

      /*
       * Now, distribute the pixel error across the adjacent pixels...
       */

       burkes_dither(p0, p1, error);

      /*
       * Update the output buffer offset & bit...
       */

       if (lbit <= 1)
        {
         lbit = 128;
         lbyte ++;
        }
       else
        {
         lbit >>= 1;
        };

      /*
       * Compute the next horizontal position...
       */

       iline  += istep;
       ierror += imod;
       if (ierror >= printer_width)
        {
         ierror -= printer_width;
         iline  += idir;
        };
      };

    /*
     * Output a line of printer graphics...
     */

     output_line(lbuf, lwidth, rgb);
    };
  };

/*
 * Deallocate all buffers used...
 */

 for (rgb = OutputType; rgb < 3; rgb ++)
  {
   free(pbuf[0][rgb]);
   free(pbuf[1][rgb]);
  };

 free(lbuf);

/*
 * End GFX and eject the current page...
 */

 if (OutputType == PCL_COLOR)
  printf("\033*rbC");
 else
  printf("\033*rB");

 printf("\033&l0H");
}


/*
 * End of "$Id: output.c,v 1.5 93/12/05 10:42:53 mike Exp Locker: mike $".
 */

